# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.readthedocs.io/
# FOGLAMP_END

""" Module for velocity filter plugin
This filter calculates velocity values from acceleration data at the original sampling rate. It also adds a datapoint
for storing acceleration value in G's. It also adds a new reading every second containing RMS and peak values for
acceleration and velocity every second.
"""

import os
import time
import copy
import logging

from foglamp.common import logger, iprpc
from foglamp.plugins.common import utils
import filter_ingest

__author__ = "Amandeep Singh Arora"
__copyright__ = "Copyright (c) 2020 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"
_LOGGER = logger.setup(__name__, level=logging.INFO)

_module_dir = os.path.dirname(__file__)

# Filter specific objects
the_callback = None
the_ingest_ref = None
the_rpc = None

G_ACCEL = 9.80665
shutdown_in_progress = False


_DEFAULT_CONFIG = {
    'plugin': {
        'description': 'Velocity filter plugin',
        'type': 'string',
        'default': 'velocity',
        'readonly': 'true'
    },
    'accelAssetName': {
        'description': 'Input asset name of reading with acceleration value',
        'type': 'string',
        'default': 'vibration',
        'displayName': 'Acceleration asset name',
        'order': "1"
    },
    'accelDatapointName': {
        'description': 'Datapoint name in reading holding the acceleration value (in m/s^2)',
        'type': 'string',
        'default': '20gpk_accelmss',
        'displayName': 'Acceleration datapoint name',
        'order': "2"
    },
    'sampling_rate': {
        'description': 'Sampling rate of acceleration data (Hz)',
        'type': 'integer',
        'default': '8000',
        'displayName': 'Sampling rate (samples per sec)',
        'order': "3"
    },
    'accelGDatapointName': {
        'description': 'Datapoint that would hold acceleration value in G',
        'type': 'string',
        'default': '20gpk_accelG',
        'displayName': 'Acceleration-G datapoint name',
        'order': "4"
    },
    'velDatapointName': {
        'description': 'Datapoint that would hold the velocity value (mm/s)',
        'type': 'string',
        'default': '20gpk_velocity_mms',
        'displayName': 'Velocity datapoint name',
        'order': "5"
    },
    'numSamplesForIntegration': {
        'description': 'Number of data points to use for integration and RMS calculation',
        'type': 'integer',
        'default': '8000',
        'displayName': 'Number of samples to use for integration',
        'minimum': '3',
        'order': "6"
    },
    'enableRmsFeatures': {
        'description': 'Whether to enable RMS features',
        'type': 'boolean',
        'default': 'true',
        'displayName': 'Enable RMS features',
        'order': "7"
    },
    'accelGRmsDatapointName': {
        'description': 'Datapoint that would hold RMS acceleration value in G',
        'type': 'string',
        'default': '20gpk_accelG',
        'displayName': 'Acceleration-G RMS datapoint name',
        'validity': "enableRmsFeatures == 'true'",
        'order': "8"
    },
    'accelGRmsPeakDatapointName': {
        'description': 'Datapoint that would hold RMS acceleration peak value in G',
        'type': 'string',
        'default': '20gpk_accelG_pk',
        'displayName': 'Acceleration-G RMS peak datapoint name',
        'validity': "enableRmsFeatures == 'true'",
        'order': "9"
    },
    'velGRmsDatapointName': {
        'description': 'Datapoint that would hold the RMS velocity value in Gs',
        'type': 'string',
        'default': '20gpk_velocity_mms',
        'displayName': 'Velocity-G RMS datapoint name',
        'validity': "enableRmsFeatures == 'true'",
        'order': "10"
    },
    'velGRmsPeakDatapointName': {
        'description': 'Datapoint that would hold the RMS velocity peak value in Gs',
        'type': 'string',
        'default': '20gpk_velocity_mms_pk',
        'displayName': 'Velocity-G RMS peak datapoint name',
        'validity': "enableRmsFeatures == 'true'",
        'order': "11"
    },
    'enableEnvelope': {
        'description': 'Whether to enable acceleration envelope output',
        'type': 'boolean',
        'default': 'true',
        'displayName': 'Enable envelope output',
        'order': "12"
    },
    'accelGEnvelopeDatapointName': {
        'description': 'Datapoint that would hold acceleration envelope value in G',
        'type': 'string',
        'default': '20gpk_accelG',
        'displayName': 'Acceleration-G envelope datapoint name',
        'validity': "enableEnvelope == 'true'",
        'order': "13"
    },
    'resonanceFreqLower': {
        'description': 'Resonance band start frequency of bearing for envelope calculation (Hz)',
        'type': 'integer',
        'default': '1000',
        'displayName': 'Bearing resonance band start frequency (Hz)',
        'minimum': '0',
        'validity': "enableEnvelope == 'true'",
        'order': "14",
    },
    'resonanceFreqUpper': {
        'description': 'Resonance band end frequency of bearing for envelope calculation (Hz)',
        'type': 'integer',
        'default': '4000',
        'displayName': 'Bearing resonance band end frequency (Hz)',
        'minimum': '0',
        'validity': "enableEnvelope == 'true'",
        'order': "15"
    },
    'enable': {
        'description': 'Enable filter plugin',
        'type': 'boolean',
        'default': 'false',
        'displayName': 'Enabled',
        'order': "16"
    },
}


def plugin_info():
    """ Returns information about the plugin
    Args:
    Returns:
        dict: plugin information
    Raises:
    """
    _LOGGER.debug("plugin_info called")
    return {
        'name': 'velocity',
        'version': '1.9.0',
        'mode': "none",
        'type': 'filter',
        'interface': '1.0',
        'config': _DEFAULT_CONFIG
    }

def parse_config(handle, config):
    """ Initialise the plugin
    Args:
        handle: handle returned by the plugin initialisation call
        config: JSON configuration document for the Filter plugin configuration category
    Returns:
        data: updated handle
    Raises:
    """
    handle['accelAssetName'] = config['accelAssetName']['value']
    handle['accelDatapointName'] = config['accelDatapointName']['value']
    handle['sampling_rate'] = int(config['sampling_rate']['value'])
    handle['enable'] = config['enable']['value'] == 'true'

    handle['accelGDatapointName'] = config['accelGDatapointName']['value']
    handle['velDatapointName'] = config['velDatapointName']['value']
    handle['numSamplesForIntegration'] = int(config['numSamplesForIntegration']['value'])

    handle['enableRmsFeatures'] = config['enableRmsFeatures']['value'] == 'true'
    handle['accelGRmsDatapointName'] = config['accelGRmsDatapointName']['value']
    handle['accelGRmsPeakDatapointName'] = config['accelGRmsPeakDatapointName']['value']
    handle['velGRmsDatapointName'] = config['velGRmsDatapointName']['value']
    handle['velGRmsPeakDatapointName'] = config['velGRmsPeakDatapointName']['value']

    handle['enableEnvelope'] = config['enableEnvelope']['value'] == 'true'
    handle['accelGEnvelopeDatapointName'] = config['accelGEnvelopeDatapointName']['value']
    handle['resonanceFreqLower'] = int(config['resonanceFreqLower']['value'])
    handle['resonanceFreqUpper'] = int(config['resonanceFreqUpper']['value'])

    return handle

def plugin_init(config, ingest_ref, callback):
    """ Initialise the plugin
    Args:
        config: JSON configuration document for the Filter plugin configuration category
        ingest_ref: Ingest reference object
        callback: Callback reference object
    Returns:
        handle: python dict object to be used in future calls to the plugin
    Raises:
    """
    global the_callback, the_ingest_ref, the_rpc
    the_callback = callback
    the_ingest_ref = ingest_ref

    # start a server for the ai algorithm the user has configured
    the_rpc = iprpc.IPCModuleClient("np_server", _module_dir)

    # all config data remains on the client side
    handle = dict()
    handle = parse_config(handle, config)
    handle['readings_buffer'] = list()
    
    return handle


def plugin_reconfigure(handle, new_config):
    """ Reconfigures the plugin
    Args:
        handle: handle returned by the plugin initialisation call
        new_config: JSON object representing the new configuration category for the category
    Returns:
        new_handle: new handle to be used in the future calls
    """
    _LOGGER.debug("Old config for velocity plugin {} \n     new config {}".format(handle, new_config))

    new_handle = copy.deepcopy(handle)  # retain readings_buffer also
    new_handle = parse_config(new_handle, new_config)  # parse new config

    return new_handle


def plugin_shutdown(handle):
    """ Shutdowns the plugin doing required cleanup.
    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
        plugin shutdown
    """
    global shutdown_in_progress, the_callback, the_ingest_ref, the_rpc
    _LOGGER.info('SHUTDOWN')
    shutdown_in_progress = True
    time.sleep(1)
    the_callback = None
    the_ingest_ref = None
    if the_rpc != None:
        try:
            the_rpc.plugin_shutdown()
        except EOFError:
            _LOGGER.info("Normal shutdown exit")
        the_rpc = None
    _LOGGER.info('Filter velocity plugin shutdown.')

    
def plugin_ingest(handle, data):
    """ Modify readings data and pass it onward
    Args:
        handle: handle returned by the plugin initialisation call
        data: readings data
    """
    global shutdown_in_progress, the_callback, the_ingest_ref, the_rpc
    
    if shutdown_in_progress:
        return

    if not handle['enable']:
        # Filter not enabled, just pass data onwards
        filter_ingest.filter_ingest_callback(the_callback, the_ingest_ref, data)
        return

    _asset_name = handle['accelAssetName']
    _datapoint_name = handle['accelDatapointName']
    _sampling_rate = handle['sampling_rate']
    _num_samples_for_integration = handle['numSamplesForIntegration']

    if type(data) == dict:
        _LOGGER.debug("data is a DICT, data={}".format(data))
        data = [data]

    if len(data) == 0:
        _LOGGER.debug("data={}".format(data))

    buffer = list()
    handle['readings_buffer'].extend(data)
    _LOGGER.debug("Got {} new readings to process, total {} in readings_buffer".format(len(data), len(handle['readings_buffer']) ))

    if len(handle['readings_buffer']) < _num_samples_for_integration:
        _LOGGER.debug("Not processing readings_buffer yet")
        return

    for reading in handle['readings_buffer'][:_num_samples_for_integration]:
        if reading['asset'] != _asset_name:
            _LOGGER.debug("Skipping reading: reading['asset']={}, _asset_name={}".format(reading['asset'], _asset_name))
            continue
        
        datapoints = reading['readings']
        if _datapoint_name not in datapoints:
            _LOGGER.debug("Skipping datapoint, {} not in {}".format(_datapoint_name, list(datapoints.keys())))
            continue
        
        accel = datapoints[_datapoint_name]
        buffer.append(accel)

    velo, accelGRms, accelGRmsPeak, velGRms, velGRmsPeak, envelope = \
        the_rpc.acc_to_velocity(buffer, _sampling_rate,
                                handle['enableEnvelope'], handle['resonanceFreqLower'], handle['resonanceFreqUpper'])

    for i,reading in enumerate(handle['readings_buffer'][:_num_samples_for_integration]):
        reading['readings'][handle['velDatapointName']] = velo[i]
        reading['readings'][handle['accelGDatapointName']]  = reading['readings'][_datapoint_name] / G_ACCEL

    time_stamp = utils.local_timestamp()

    if handle['enableRmsFeatures']:
        rms_reading = {'asset': _asset_name+'_RMS', 'timestamp': time_stamp, 'readings': {} }

        rms_reading['readings'][handle['accelGRmsDatapointName']] = accelGRms
        rms_reading['readings'][handle['accelGRmsPeakDatapointName']] = accelGRmsPeak
        rms_reading['readings'][handle['velGRmsDatapointName']] = velGRms
        rms_reading['readings'][handle['velGRmsPeakDatapointName']] = velGRmsPeak

        _LOGGER.debug("accelGRms={:.4f}, velGRms={:.4f}".format(accelGRms, velGRms))

    envelope_reading = {'asset': _asset_name + '_envelope', 'timestamp': time_stamp, 'readings': {}}
    envelope_reading['readings'][handle['accelGEnvelopeDatapointName']] = envelope

    _LOGGER.debug("Updated {} readings to include RMS/peak acceleration/velocity/accel-envelope values".format(_num_samples_for_integration))

    if True:
        readings_buffer_to_fwd = handle['readings_buffer'][:_num_samples_for_integration]
    else:
        readings_buffer_to_fwd = []  # to avoid filling disk with raw data at high rate

    if handle['enableRmsFeatures']:
        readings_buffer_to_fwd.append(rms_reading)

    if handle['enableEnvelope']:
        readings_buffer_to_fwd.append(envelope_reading)

    filter_ingest.filter_ingest_callback(the_callback, the_ingest_ref, readings_buffer_to_fwd)

    handle['readings_buffer'] = handle['readings_buffer'][_num_samples_for_integration:]
    # _LOGGER.debug("{} readings left in readings_buffer".format(len(handle['readings_buffer'])))
